import math, random

import gym
import copy
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.font_manager

import sys
import torch
import torch.nn as nn
import torch.optim as optim
import torch.autograd as autograd 
import torch.nn.functional as F
from torch.utils.data import Dataset
from torch.utils.data import DataLoader
from itertools import cycle

import sys
sys.path.append("..")
sys.path.append("../..")

import general_energy_nets.predictive_coding as pc

from value_iteration import value_iteration, plot_grid

from IPython.display import clear_output

import argparse

ENV_ID = "FrozenLake-v1"

PC_LR = 0.001
PC_OPTIM = 'Adam'

BP_LR = 0.001
BP_OPTIM = 'Adam'

BATCH_SIZE = 100
GAMMA = 0.7
MAX_ITERS = 100

class QValueDataset(Dataset):
    
    def __init__(self, target_state_values, gamma=0.7):
        self.state_action_values = []
        
        for state in range(env.nS):
            targets = np.zeros(env.nA)
            for action in range(env.nA):
                _, next_state, reward, done = env.P[state][action][0]
                q_value = reward + gamma*target_state_values[next_state]            
                targets[action] = q_value
            self.state_action_values.append((state, targets))
                
    def __len__(self):
        return len(self.state_action_values)

    def __getitem__(self, idx):
        state, targets = self.state_action_values[idx]
        return F.one_hot(torch.tensor(state).to(torch.int64), num_classes=env.observation_space.n).to(torch.float32), torch.FloatTensor(targets)


class DQN(nn.Module):
    def __init__(self, num_inputs, num_actions, hidden_size=128, predictive_coding=False):
        super(DQN, self).__init__()
        
        self.layers = []
        self.layers.append(nn.Linear(num_inputs, hidden_size))
        if predictive_coding:
            self.layers.append(pc.PCLayer())
        self.layers.append(nn.ReLU())
        self.layers.append(nn.Linear(hidden_size, hidden_size))
        if predictive_coding:
            self.layers.append(pc.PCLayer())
        self.layers.append(nn.ReLU())
        self.layers.append(nn.Linear(hidden_size, num_actions))
        self.layers = nn.Sequential(*self.layers)
        
    def forward(self, x):
        return self.layers(x)
    
    @torch.no_grad()
    def get_policy_values(self):
        V = np.zeros(env.nS)
        policy = np.zeros([env.nS, env.nA]) / env.nA

        for s in range(env.nS):
            _s = F.one_hot(torch.tensor(s), num_classes=env.observation_space.n).to(torch.float32).unsqueeze(0).to(device)
            q_value = self.forward(_s)
            V[s] = q_value.max().cpu().item()
            policy[s] = F.one_hot(q_value.argmax().cpu(), num_classes=env.action_space.n).to(torch.float32)
            
        return policy, V

def loss_fn(output, _targets):
    return (output - _targets).pow(2).sum() * 0.5

def step_update(net, trainer, state, action):
    net.train()
    
    results = trainer.train_on_batch(
        inputs=state,
        loss_fn=loss_fn,
        loss_fn_kwargs={
            '_targets': action
        },
        is_log_progress=False,
        is_return_results_every_t=False,
        is_checking_after_callback_after_t=False,
    )
    
    return results

def compute_policy_loss(optimal_policy, policy):
    # compute 0-1 policy loss
    policy_loss = 0.0
    
    for s in range(env.nS):
        # if the policy doesn't point in the same direction then it is not optimal
        if np.dot(policy[s], optimal_policy[s]) == 0.0:
            policy_loss += 1.0

    return policy_loss

def compute_value_loss(optimal_V, V):
    # compute mse value loss
    value_loss = 0.0
        
    return ((optimal_V - V)**2).sum()

def plot(bp_policy_loss, bp_value_loss, pc_policy_loss, pc_value_loss):
    plt.style.use('ggplot')
    max_iters = MAX_ITERS

    plt.figure(figsize=(20,5))
    plt.subplot(131)
    plt.title('BP 0-1 policy losses')
    plt.plot([x for x in range(len(bp_policy_loss))], bp_policy_loss, linewidth=2, color='blueviolet', alpha=0.8, label="BP")

    plt.subplot(132)
    plt.title('BP MSE value losses')
    plt.plot([x for x in range(len(bp_value_loss))], bp_value_loss, linewidth=2, color='blueviolet', alpha=0.8, label="BP")

    plt.savefig('./imgs/supervised_{}_bp_lr={}_bp_optim={}.png'.format(ENV_ID, BP_LR, BP_OPTIM))
    plt.close('all')

    plt.figure(figsize=(20,5))
    plt.subplot(131)
    plt.title('PC 0-1 policy losses')
    plt.plot([x for x in range(len(pc_policy_loss))], pc_policy_loss, linewidth=2, color='firebrick', alpha=0.8, label="PC")

    plt.subplot(132)
    plt.title('PC MSE value losses')
    plt.plot([x for x in range(len(pc_value_loss))], pc_value_loss, linewidth=2, color='firebrick', alpha=0.8, label="PC")

    plt.savefig('./imgs/supervised_{}_pc_lr={}_pc_optim={}.png'.format(ENV_ID, PC_LR, PC_OPTIM))
    plt.close('all')

    plt.figure(figsize=(20,5))
    plt.subplot(131)
    plt.title('0-1 policy losses')
    plt.plot([x for x in range(len(pc_policy_loss))], pc_policy_loss, linewidth=2, color='firebrick', alpha=0.8, label="PC")
    plt.plot([x for x in range(len(bp_policy_loss))], bp_policy_loss, linewidth=2, color='blueviolet', alpha=0.8, label="BP")
    plt.legend()

    plt.subplot(132)
    plt.title('MSE value losses')
    plt.plot([x for x in range(len(pc_value_loss))], pc_value_loss, linewidth=2, color='firebrick', alpha=0.8, label="PC")
    plt.plot([x for x in range(len(bp_value_loss))], bp_value_loss, linewidth=2, color='blueviolet', alpha=0.8, label="BP")
    plt.legend()

    plt.savefig('./imgs/supervised_{}_bp_lr={}_bp_optim={}_pc_lr={}_pc_lr={}.png'.format(ENV_ID, BP_LR, BP_OPTIM, PC_LR, PC_OPTIM))
    plt.close('all')

def save_data(bp_policy_loss_matrix, pc_policy_loss_matrix, bp_value_loss_matrix, pc_value_loss_matrix):
    bp_mean_policy_loss = np.mean(bp_policy_loss_matrix)
    pc_mean_policy_loss = np.mean(pc_policy_loss_matrix)

    bp_mean_value_loss = np.mean(bp_value_loss_matrix)
    pc_mean_value_loss = np.mean(pc_value_loss_matrix)

    with open('./data/supervised_{}_bp_lr={}_bp_optim={}_pc_lr={}_pc_lr={}.txt'.format(ENV_ID, BP_LR, BP_OPTIM, PC_LR, PC_OPTIM), 'w') as outfile:
        outfile.write('BP Mean Policy Loss ({}) : \n'.format(np.mean(bp_mean_policy_loss)))
        outfile.write(str(bp_mean_policy_loss))
        outfile.write('\n')
        outfile.write('PC Mean Policy Loss ({}) : \n'.format(np.mean(pc_mean_policy_loss)))
        outfile.write(str(pc_mean_policy_loss))
        outfile.write('\n')
        outfile.write('BP Mean Value Loss ({}) : \n'.format(np.mean(bp_mean_value_loss)))
        outfile.write(str(bp_mean_value_loss))
        outfile.write('\n')
        outfile.write('PC Mean Value Loss ({}) : \n'.format(np.mean(pc_mean_value_loss)))
        outfile.write(str(pc_mean_value_loss))
        outfile.write('\n')


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--env-id", default=ENV_ID)
    parser.add_argument("--batch-size", default=BATCH_SIZE)
    parser.add_argument("--bp-lr", default=BP_LR)
    parser.add_argument("--pc-lr", default=PC_LR)
    parser.add_argument("--bp-optim", default=BP_OPTIM)
    parser.add_argument("--pc-optim", default=PC_OPTIM)
    parser.add_argument("--max-iters", default=MAX_ITERS)
    args = parser.parse_args()

    ENV_ID = args.env_id
    BATCH_SIZE = int(args.batch_size)
    BP_LR = float(args.bp_lr)
    PC_LR = float(args.pc_lr)
    BP_OPTIM = args.bp_optim
    PC_OPTIM = args.pc_optim
    MAX_ITERS = int(args.max_iters)

    assert torch.cuda.is_available()
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu" )

    env = gym.make(ENV_ID, is_slippery=False) 

    seed = 745
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)
    env.action_space.seed(seed)
    env.seed(seed)

    policy_optim, V_optim = value_iteration(env, gamma=GAMMA)
    policy_optim = torch.FloatTensor(policy_optim)

    dataset = QValueDataset(V_optim, GAMMA)

    data_loader = DataLoader(dataset, batch_size=BATCH_SIZE, shuffle=True)
    data_iter = cycle(data_loader)

    # run BP

    model = DQN(env.observation_space.n, env.action_space.n, predictive_coding=False).to(device)
    model.eval()

    optimizer = None
    if BP_OPTIM == 'Adam':
        optimizer = optim.Adam
    elif BP_OPTIM == 'SGD':
        optimizer = optim.SGD   
    else:
        raise ValueError("Bad argument for optim")

    bp_trainer = pc.PCTrainer(
        model,
        T=1,
        update_x_at='all',
        optimizer_x_fn=optim.SGD,
        optimizer_x_kwargs={"lr": 0.1},
        x_lr_discount=0.5,
        update_p_at='last',
        optimizer_p_fn=optimizer,
        optimizer_p_kwargs={"lr": BP_LR},
        plot_progress_at=[],
    )

    bp_policy_losses = []
    bp_value_losses = []
    losses = []

    for iteration in range(1, MAX_ITERS + 1):
        state, targets = next(data_iter)
        state, targets = state.to(device), targets.to(device)
        
        results = step_update(model, bp_trainer, state, targets)
            
        loss = results['loss'][0]
        losses.append(loss)
                
        model.eval()
        policy, V = model.get_policy_values()
        policy_loss = compute_policy_loss(policy_optim, policy)
        value_loss = compute_value_loss(V_optim, V)
        bp_policy_losses.append(policy_loss)
        bp_value_losses.append(value_loss)
                
        if (iteration == MAX_ITERS):
            print("Iteration {} batch loss {} policy loss {} value loss {}".format(
                iteration, np.mean(losses[-10:]), np.mean(bp_policy_losses[-10:]), np.mean(bp_value_losses[-10:])
            ))
            break

    # run PC

    model = DQN(env.observation_space.n, env.action_space.n, predictive_coding=True).to(device)
    model.eval()

    optimizer = None
    if PC_OPTIM == 'Adam':
        optimizer = optim.Adam
    elif PC_OPTIM == 'SGD':
        optimizer = optim.SGD   
    else:
        raise ValueError("Bad argument for optim")

    pc_trainer = pc.PCTrainer(
        model,
        T=128,
        update_x_at='all',
        optimizer_x_fn=optim.SGD,
        optimizer_x_kwargs={"lr": 0.1},
        x_lr_discount=0.5,
        update_p_at='last',
        optimizer_p_fn=optimizer,
        optimizer_p_kwargs={"lr": PC_LR},
        plot_progress_at=[],
    )

    pc_policy_losses = []
    pc_value_losses = []
    losses = []

    for iteration in range(1, MAX_ITERS + 1):
        state, targets = next(data_iter)
        state, targets = state.to(device), targets.to(device)
        
        results = step_update(model, pc_trainer, state, targets)
            
        loss = results['loss'][0]
        losses.append(loss)
                
        model.eval()
        policy, V = model.get_policy_values()
        policy_loss = compute_policy_loss(policy_optim, policy)
        value_loss = compute_value_loss(V_optim, V)
        pc_policy_losses.append(policy_loss)
        pc_value_losses.append(value_loss)
                
        if (iteration == MAX_ITERS):
            print("Iteration {} batch loss {} policy loss {} value loss {}".format(
                iteration, np.mean(losses[-10:]), np.mean(pc_policy_losses[-10:]), np.mean(pc_value_losses[-10:])
            ))
            break

    plot(bp_policy_losses, bp_value_losses, pc_policy_losses, pc_value_losses)
    save_data(bp_policy_losses, pc_policy_losses, bp_value_losses, pc_value_losses)