import math
import time
import gym
import numpy as np
import pybullet_envs

import sys
import torch
import torch.nn as nn
import torch.optim as optim
import torch.autograd as autograd 
import torch.nn.functional as F

import matplotlib.pyplot as plt

import sys
sys.path.append("..")
sys.path.append("../..")

import general_energy_nets.predictive_coding as pc

import argparse

NUM_SEEDS = 1
ENV_ID = "AntBulletEnv-v0"
GAMMA = 0.99
GAE_LAMBDA = 0.95

TRAJECTORY_SIZE = 2049
LEARNING_RATE_ACTOR = 0.00001
LEARNING_RATE_CRITIC = 0.0001

PPO_EPS = 0.2
PPO_EPOCHES = 10
PPO_BATCH_SIZE = 64

ENTROPY_BETA = 0.0

HID_SIZE = 64

TEST_ITERS = 100000

MAX_FRAMES = 20000000

SCALE_FACTOR = 1.0

class Actor(nn.Module):
    def __init__(self, obs_size, act_size):
        super(Actor, self).__init__()

        self.mu = nn.Sequential(
            nn.Linear(obs_size, HID_SIZE),
            pc.PCLayer(),
            nn.Tanh(),
            nn.Linear(HID_SIZE, HID_SIZE),
            pc.PCLayer(),
            nn.Tanh(),
            nn.Linear(HID_SIZE, act_size),
            pc.PCLayer(),
            nn.Tanh(),
        )
        
        self.logstd = nn.Parameter(torch.zeros(act_size))

    def forward(self, x):
        return self.mu(x)


class Critic(nn.Module):
    def __init__(self, obs_size):
        super(Critic, self).__init__()

        self.value = nn.Sequential(
            nn.Linear(obs_size, HID_SIZE),
            pc.PCLayer(),
            nn.ReLU(),
            nn.Linear(HID_SIZE, HID_SIZE),
            pc.PCLayer(),
            nn.ReLU(),
            nn.Linear(HID_SIZE, 1),
        )

    def forward(self, x):
        return self.value(x)

def test_net(net, env, count=10, device="cpu"):
    rewards = 0.0
    steps = 0
    for _ in range(count):
        state = env.reset()
        while True:
            state = torch.tensor(np.array(state, dtype=np.float32)).to(device)
            mu = actor(state).data.cpu().numpy()
            action = np.clip(mu, -1, 1)
            state, reward, done, _ = env.step(action)
            rewards += reward
            steps += 1
            if done:
                break
    return rewards / count, steps / count

def calc_adv_ref(values, dones, rewards, gamma, gae_lambda):
    last_gae = 0.0
    adv, ref = [], []

    for val, next_val, done, reward in zip(reversed(values[:-1]), reversed(values[1:]),
                                           reversed(dones[:-1]), reversed(rewards[:-1])):
        if done:
            delta = reward - val
            last_gae = delta
        else:
            delta = reward + gamma * next_val - val
            last_gae = delta + gamma * gae_lambda * last_gae
        adv.append(last_gae)
        ref.append(last_gae + val)
        
    adv = list(reversed(adv))
    ref = list(reversed(ref))
    return torch.FloatTensor(adv), torch.FloatTensor(ref)

def calc_logprob(mu_v, logstd_v, actions_v):
    p1 = - ((mu_v - actions_v) ** 2) / (2*torch.exp(logstd_v).clamp(min=1e-3))
    p2 = - torch.log(torch.sqrt(2 * math.pi * torch.exp(logstd_v)))
    return p1 + p2

def calc_entropy(logstd_v, actions_v):
    return (torch.zeros_like(actions_v) + 0.5 * (1 + torch.log(2 * torch.pi * torch.exp(logstd_v)**2))).sum(dim=1).mean()   

def train(batch):
    states_t, actions_t, advantage, ref, old_logprob = batch
    advantage = advantage.unsqueeze(-1)
    
    critic.train()
    results = critic_trainer.train_on_batch(
        inputs=states_t,
        loss_fn=critic_loss_fn,
        loss_fn_kwargs={
            'ref': ref
        },
        is_log_progress=False,
        is_return_results_every_t=False,
        is_checking_after_callback_after_t=False,
    )
    
    actor.train()
    results = actor_trainer.train_on_batch(
        inputs=states_t,
        loss_fn=actor_loss_fn,
        loss_fn_kwargs={
            'actions': actions_t,
            'old_logprob': old_logprob,
            'advantage': advantage
        },
        is_log_progress=False,
        is_return_results_every_t=False,
        is_checking_after_callback_after_t=False,
    )

def critic_loss_fn(output, ref):
    return (output.squeeze(-1) - ref).pow(2).sum() * 0.5
    #return F.mse_loss(output.squeeze(-1), ref) * CRITIC_SCALE_FACTOR 

def actor_loss_fn(output, actions, old_logprob, advantage):
    entropy = calc_entropy(actor.logstd, actions_t)

    logprob = calc_logprob(output, actor.logstd, actions) 
    ratio = torch.exp(logprob - old_logprob)
    surr_obj = advantage * ratio
    clipped_surr = advantage * torch.clamp(ratio, 1.0 - PPO_EPS, 1.0 + PPO_EPS)
    return -torch.min(surr_obj, clipped_surr).sum() * SCALE_FACTOR - ENTROPY_BETA * entropy

def plot(reward_matrix, confidence_intervals=False):
    plt.style.use('ggplot')
    average_reward = np.mean(np.mean(reward_matrix, axis=0))
    plt.figure(figsize=(10, 5))
    plt.title('{} : PPO with PC (Avg Reward = {:.2f})'.format(ENV_ID, average_reward))
    reward_matrix_mean = np.mean(reward_matrix, axis=0)
    plt.plot([x * TEST_ITERS for x in range(reward_matrix.shape[1])], reward_matrix_mean, linewidth=2, color='firebrick', alpha=0.8)

    if confidence_intervals:
        upper_confidence_interval = []
        lower_confidence_interval = []

        for episode in reward_matrix.transpose():
            bootstrapped_means = []
            for B in range(2000):
                bootstrapped_sample = np.random.choice(episode, len(episode))
                bootstrapped_means.append(np.mean(bootstrapped_sample))
            bootstrapped_percentiles = np.percentile(np.array(bootstrapped_means), [2.5, 97.5])
            upper_confidence_interval.append(bootstrapped_percentiles[1])
            lower_confidence_interval.append(bootstrapped_percentiles[0])

        plt.fill_between([x * TEST_ITERS for x in range(reward_matrix.shape[1])], upper_confidence_interval, lower_confidence_interval, color='firebrick', alpha=0.4)

    plt.xlabel('frames')
    plt.ylabel('test reward')
    plt.savefig('./imgs/PPO_pc_{}_seeds={}_lrc={}_lra={}_sf={}.png'.format(ENV_ID, NUM_SEEDS, LEARNING_RATE_CRITIC, LEARNING_RATE_ACTOR, SCALE_FACTOR))
    plt.close('all')

def save(reward_matrix):
    with open('./data/PPO_pc_{}_seeds={}_lrc={}_lra={}_sf={}.npy'.format(ENV_ID, NUM_SEEDS, LEARNING_RATE_CRITIC, LEARNING_RATE_ACTOR, SCALE_FACTOR), 'wb') as outfile:
        np.save(outfile, reward_matrix)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--env-id", default=ENV_ID)
    parser.add_argument("--n-seed", default=NUM_SEEDS)
    parser.add_argument("--lrc", default=LEARNING_RATE_CRITIC)
    parser.add_argument("--lra", default=LEARNING_RATE_ACTOR)
    parser.add_argument("--max-frames", default=MAX_FRAMES)
    parser.add_argument("--scale-factor", default=SCALE_FACTOR)
    parser.add_argument("--beta", default=ENTROPY_BETA)
    args = parser.parse_args()

    ENV_ID = args.env_id
    NUM_SEEDS = int(args.n_seed)
    LEARNING_RATE_CRITIC = float(args.lrc)
    LEARNING_RATE_ACTOR = float(args.lra)
    MAX_FRAMES = int(args.max_frames)
    SCALE_FACTOR = float(args.scale_factor)
    ENTROPY_BETA = float(args.beta)

    assert torch.cuda.is_available()
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu" )

    reward_matrix = []

    for SEED in range(1, NUM_SEEDS+1):
        env = gym.make(ENV_ID)
        test_env = gym.make(ENV_ID)

        torch.manual_seed(SEED)
        np.random.seed(SEED)
        env.action_space.seed(SEED)
        env.seed(SEED)

        test_env.action_space.seed(SEED+1)
        test_env.seed(SEED+1)

        actor = Actor(env.observation_space.shape[0], env.action_space.shape[0]).to(device)
        critic = Critic(env.observation_space.shape[0]).to(device)

        actor_trainer = pc.PCTrainer(
            actor,
            T=32,
            update_x_at='all',
            optimizer_x_fn=optim.SGD,
            optimizer_x_kwargs={"lr": 0.05},
            x_lr_discount=0.5,
            update_p_at='last',
            optimizer_p_fn=optim.Adam,
            optimizer_p_kwargs={"lr": LEARNING_RATE_ACTOR},
            plot_progress_at=[],
        )

        critic_trainer = pc.PCTrainer(
            critic,
            T=32,
            update_x_at='all',
            optimizer_x_fn=optim.SGD,
            optimizer_x_kwargs={"lr": 0.05},
            x_lr_discount=0.5,
            update_p_at='last',
            optimizer_p_fn=optim.Adam,
            optimizer_p_kwargs={"lr": LEARNING_RATE_CRITIC},
            plot_progress_at=[],
        )

        frame_idx = 0
        ts = time.time()

        all_mean_rewards = []

        all_rewards = []
        episode_reward = 0.0

        while frame_idx < MAX_FRAMES:
            states = []
            actions = []
            rewards = []
            dones = []
            last_done_index = None
            
            state = env.reset()
            for _ in range(TRAJECTORY_SIZE):
                with torch.no_grad():
                    actor.eval()
                    state = torch.tensor(np.array(state, dtype=np.float32)).to(device)
                    mu = actor(state).data.cpu().numpy()
                    logstd = actor.logstd.data.cpu().numpy()
                    rnd = np.random.normal(size=logstd.shape)
                    action = mu + np.exp(logstd) * rnd
                    action = np.clip(action, -1, 1)

                    next_state, reward, done, _ = env.step(action)

                    episode_reward += reward

                    states.append(state)
                    actions.append(torch.tensor(action))
                    rewards.append(reward)
                    dones.append(done)

                    state = next_state

                    if done:
                        all_rewards.append(episode_reward)
                        last_done_index = len(states)-1
                        state = env.reset()
                        episode_reward = 0

                frame_idx += 1
                
                if frame_idx % TEST_ITERS == 0:
                    print("Test (frame %s): %.2f frames per sec, reward %.3f" % (frame_idx, TEST_ITERS / (time.time() - ts), np.mean(all_rewards[-100:])))
                    ts = time.time()
                    all_mean_rewards.append(np.mean(all_rewards[-100:]))
                
                if len(states) < TRAJECTORY_SIZE:
                    continue
                    
                if last_done_index is None or last_done_index == len(states)-1:
                    continue
                    
                states = states[:last_done_index+2]
                actions = actions[:last_done_index + 2]
                rewards = rewards[:last_done_index + 2]
                dones = dones[:last_done_index + 2]
                
                states_t = torch.stack(states)
                actions_t = torch.stack(actions).to(device)
                
                critic.eval()
                values = critic(states_t)
                values = values.squeeze().data.cpu().numpy()

                advantage, ref = calc_adv_ref(values, dones, rewards, GAMMA, GAE_LAMBDA)
                advantage = advantage.to(device)
                ref = ref.to(device)
                
                actor.eval()
                mu = actor(states_t)
                old_logprob = calc_logprob(mu, actor.logstd, actions_t)
                advantage = (advantage - torch.mean(advantage)) / torch.std(advantage)
                old_logprob = old_logprob.detach()

                # make our trajectory splittable on even batch chunks
                len_trajectory = len(states_t) - 1
                len_trajectory -= len_trajectory % PPO_BATCH_SIZE
                len_trajectory += 1
                indices = np.arange(0, len_trajectory-1)
                
                for _ in range(PPO_EPOCHES):
                    np.random.shuffle(indices)
                    for batch_indices in np.split(indices, len_trajectory // PPO_BATCH_SIZE):
                        train((states_t[batch_indices],
                            actions_t[batch_indices],
                            advantage[batch_indices],
                            ref[batch_indices],
                            old_logprob[batch_indices]
                            ))
                    
                states.clear()
                actions.clear()
                rewards.clear()
                dones.clear()

        reward_matrix.append(all_mean_rewards)

    reward_matrix = np.array(reward_matrix)
    plot(reward_matrix, confidence_intervals=(NUM_SEEDS >= 3))
    save(reward_matrix)